/*arch/i386/kernel/sys_i386.c
 *
 * This file contains various random system calls that
 * have a non-standard calling sequence on the Linux/i386
 * platform.
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <linux/shm.h>
#include <linux/stat.h>
#include <linux/mman.h>
#include <linux/file.h>
#include <linux/utsname.h>
#include <linux/module.h>

#include <asm/uaccess.h>
#include <asm/ipc.h>
#include <asm/io.h>
#include <linux/slab.h>

//here are some constants used with fast guest context switching
#define NUM_PGD_GUEST 448
#define VMA_LIST_IDX  1023
#define GUEST_APP_LIMIT 0x70000000
#define PTRS_PER_PAGE 1024

int stkprint=0;
EXPORT_SYMBOL_NOVERS(stkprint);


/*
 * sys_pipe() is the normal C calling standard for creating
 * a pipe. It's not the way Unix traditionally does this, though.
 */
asmlinkage int sys_pipe(unsigned long * fildes)
{
	int fd[2];
	int error;

	error = do_pipe(fd);
	if (!error) {
		if (copy_to_user(fildes, fd, 2*sizeof(int)))
			error = -EFAULT;
	}
	return error;
}

/* common code for old and new mmaps */
static inline long do_mmap2(
	unsigned long addr, unsigned long len,
	unsigned long prot, unsigned long flags,
	unsigned long fd, unsigned long pgoff)
{
	int error = -EBADF;
	struct file * file = NULL;

	flags &= ~(MAP_EXECUTABLE | MAP_DENYWRITE);
	if (!(flags & MAP_ANONYMOUS)) {
		file = fget(fd);
		if (!file)
			goto out;
	}
	down_write(&current->mm->mmap_sem);
	error = do_mmap_pgoff(file, addr, len, prot, flags, pgoff);
	up_write(&current->mm->mmap_sem);
	if (file)
		fput(file);
out:
	return error;
}

asmlinkage long sys_mmap2(unsigned long addr, unsigned long len,
	unsigned long prot, unsigned long flags,
	unsigned long fd, unsigned long pgoff)
{
	return do_mmap2(addr, len, prot, flags, fd, pgoff);
}

EXPORT_SYMBOL_NOVERS(sys_mmap2);

/*
 * Perform the select(nd, in, out, ex, tv) and mmap() system
 * calls. Linux/i386 didn't use to be able to handle more than
 * 4 system call parameters, so these system calls used a memory
 * block for parameter passing..
 */

struct mmap_arg_struct {
	unsigned long addr;
	unsigned long len;
	unsigned long prot;
	unsigned long flags;
	unsigned long fd;
	unsigned long offset;
};

asmlinkage int old_mmap(struct mmap_arg_struct *arg)
{
	struct mmap_arg_struct a;
	int err = -EFAULT;

	if (copy_from_user(&a, arg, sizeof(a)))
		goto out;

	err = -EINVAL;
	if (a.offset & ~PAGE_MASK)
		goto out;

	err = do_mmap2(a.addr, a.len, a.prot, a.flags, a.fd, a.offset >> PAGE_SHIFT);
out:
	return err;
}


extern asmlinkage int sys_select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

struct sel_arg_struct {
	unsigned long n;
	fd_set *inp, *outp, *exp;
	struct timeval *tvp;
};

asmlinkage int old_select(struct sel_arg_struct *arg)
{
	struct sel_arg_struct a;

	if (copy_from_user(&a, arg, sizeof(a)))
		return -EFAULT;
	/* sys_select() does the appropriate kernel locking */
	return sys_select(a.n, a.inp, a.outp, a.exp, a.tvp);
}

/*
 * sys_ipc() is the de-multiplexer for the SysV IPC calls..
 *
 * This is really horribly ugly.
 */
asmlinkage int sys_ipc (uint call, int first, int second,
			int third, void *ptr, long fifth)
{
	int version, ret;

	version = call >> 16; /* hack for backward compatibility */
	call &= 0xffff;

	switch (call) {
	case SEMOP:
		return sys_semop (first, (struct sembuf *)ptr, second);
	case SEMGET:
		return sys_semget (first, second, third);
	case SEMCTL: {
		union semun fourth;
		if (!ptr)
			return -EINVAL;
		if (get_user(fourth.__pad, (void **) ptr))
			return -EFAULT;
		return sys_semctl (first, second, third, fourth);
	}

	case MSGSND:
		return sys_msgsnd (first, (struct msgbuf *) ptr, 
				   second, third);
	case MSGRCV:
		switch (version) {
		case 0: {
			struct ipc_kludge tmp;
			if (!ptr)
				return -EINVAL;
			
			if (copy_from_user(&tmp,
					   (struct ipc_kludge *) ptr, 
					   sizeof (tmp)))
				return -EFAULT;
			return sys_msgrcv (first, tmp.msgp, second,
					   tmp.msgtyp, third);
		}
		default:
			return sys_msgrcv (first,
					   (struct msgbuf *) ptr,
					   second, fifth, third);
		}
	case MSGGET:
		return sys_msgget ((key_t) first, second);
	case MSGCTL:
		return sys_msgctl (first, second, (struct msqid_ds *) ptr);

	case SHMAT:
		switch (version) {
		default: {
			ulong raddr;
			ret = sys_shmat (first, (char *) ptr, second, &raddr);
			if (ret)
				return ret;
			return put_user (raddr, (ulong *) third);
		}
		case 1:	/* iBCS2 emulator entry point */
			if (!segment_eq(get_fs(), get_ds()))
				return -EINVAL;
			return sys_shmat (first, (char *) ptr, second, (ulong *) third);
		}
	case SHMDT: 
		return sys_shmdt ((char *)ptr);
	case SHMGET:
		return sys_shmget (first, second, third);
	case SHMCTL:
		return sys_shmctl (first, second,
				   (struct shmid_ds *) ptr);
	default:
		return -EINVAL;
	}
}

/*
 * Old cruft
 */
asmlinkage int sys_uname(struct old_utsname * name)
{
	int err;
	if (!name)
		return -EFAULT;
	down_read(&uts_sem);
	err=copy_to_user(name, &system_utsname, sizeof (*name));
	up_read(&uts_sem);
	return err?-EFAULT:0;
}

asmlinkage int sys_olduname(struct oldold_utsname * name)
{
	int error;

	if (!name)
		return -EFAULT;
	if (!access_ok(VERIFY_WRITE,name,sizeof(struct oldold_utsname)))
		return -EFAULT;
  
  	down_read(&uts_sem);
	
	error = __copy_to_user(&name->sysname,&system_utsname.sysname,__OLD_UTS_LEN);
	error |= __put_user(0,name->sysname+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->nodename,&system_utsname.nodename,__OLD_UTS_LEN);
	error |= __put_user(0,name->nodename+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->release,&system_utsname.release,__OLD_UTS_LEN);
	error |= __put_user(0,name->release+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->version,&system_utsname.version,__OLD_UTS_LEN);
	error |= __put_user(0,name->version+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->machine,&system_utsname.machine,__OLD_UTS_LEN);
	error |= __put_user(0,name->machine+__OLD_UTS_LEN);
	
	up_read(&uts_sem);
	
	error = error ? -EFAULT : 0;

	return error;
}

asmlinkage int sys_pause(void)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule();
	return -ERESTARTNOHAND;
}


static void printVmaList(struct vm_area_struct *vma) {
  char printBuf[100];
  while(vma != NULL) {
    sprintf(printBuf,"vma 0x%08lx, vm_start 0x%08lx, vm_end 0x%08lx\n",
	    (unsigned long) vma, vma->vm_start, vma->vm_end);
    outputStr(printBuf);
    vma = vma->vm_next;
  }
}

/*
static int freeVmaList(struct vm_area_struct *vma) {
  struct vm_area_struct *nextVma;
  int count=0;
  while(vma != NULL) {
    nextVma = vma->vm_next;
    kmem_cache_free(vm_area_cachep,vma);
    vma = nextVma;
    if(count++ > 100000) {
      BUG();
    }
  }

  return count;
}  
*/

/* Remove one vm structure from the inode's i_mapping address space. */
static inline void __remove_shared_vm_struct(struct vm_area_struct *vma)
{
	struct file * file = vma->vm_file;

	if (file) {
		struct inode *inode = file->f_dentry->d_inode;
		if (vma->vm_flags & VM_DENYWRITE)
			atomic_inc(&inode->i_writecount);
		if(vma->vm_next_share)
			vma->vm_next_share->vm_pprev_share = vma->vm_pprev_share;
		*vma->vm_pprev_share = vma->vm_next_share;
	}
}

static inline void remove_shared_vm_struct(struct vm_area_struct *vma)
{
	lock_vma_mappings(vma);
	__remove_shared_vm_struct(vma);
	unlock_vma_mappings(vma);
}

extern asmlinkage long sys_munmap(unsigned long addr, size_t len);

static struct vm_area_struct *removeVmaRange(struct mm_struct *mm,
					     unsigned long endAddress,
					     int *numRemoved) {
  int count=0;
  struct vm_area_struct *removedVmaList = NULL;
  struct vm_area_struct *vma = NULL;

  if(mm != NULL) {
    vma = mm->mmap;
  }

  if(vma != NULL) {
    if(vma->vm_start < endAddress) {

      // remove the head
      count=1;
      removedVmaList = vma;
      remove_shared_vm_struct(vma);
      while(vma->vm_next != NULL) {
	// go through the list (it is order) and remove items
	if(vma->vm_next->vm_start >= endAddress) {
	  break;
	}

	if(vma->vm_end > endAddress) {
	  BUG();
	}

	vma = vma->vm_next;
	remove_shared_vm_struct(vma);
	count++;
      }
   
      // fix up the mm struct
      mm->mmap = vma->vm_next;
      vma->vm_next = NULL;
      mm->map_count -= count;

      // reset the cache and the rb tree so we can insert if needed
      mm->mmap_cache = mm->mmap;
      build_mmap_rb(mm);

    } 
  }

  if(numRemoved != NULL) {
    *numRemoved = count;
  }

  return removedVmaList;
}  

static void insertVmaList(struct mm_struct *mm, struct vm_area_struct *vma) {
  struct vm_area_struct *next;
  int count=0;
 
  while(vma) {
    next = vma->vm_next;
    
    // I am pretty sure this takes care of counts and other stuff.  
    // revisit this later
    lock_vma_mappings(vma);
    __insert_vm_struct(mm,vma);
    unlock_vma_mappings(vma);
    vma = next;
    if(count++ > 100000) {
      BUG();
    }
  }
}

/*
static void freePgdPages(unsigned long *pgd) {
  int i;
  //free the second level page tables (we can cheat because we know it is
  // only two levels
  for(i=0;i<NUM_PGD_GUEST;i++) {
    if(pgd[i]) {
      unsigned long *pte = (unsigned long *) (__va(pgd[i] & PAGE_MASK));
      free_page((unsigned long) pte);
    }
  }
}
*/

static void do_switchguest(unsigned long *prevPgd, unsigned long *nextPgd) {
  int count=0,i;
  struct vm_area_struct *newPrevVmaList,*nextVma;
  unsigned long pgdin[NUM_PGD_GUEST], pgdout[NUM_PGD_GUEST];

  if(nextPgd != NULL) {
    unsigned long *nextVmaList = nextPgd + VMA_LIST_IDX;
    memcpy(&nextVma,nextVmaList,sizeof(unsigned long));
    memcpy(pgdin,nextPgd,NUM_PGD_GUEST*sizeof(unsigned long));     
    
    // get rid of the reference to the vma list
    nextPgd[VMA_LIST_IDX] = 0;
    for(i=0;i<NUM_PGD_GUEST;i++) {
      nextPgd[i] = 0;
    }
  } else {
      nextVma = NULL;
      memset(pgdin,0,NUM_PGD_GUEST*sizeof(unsigned long));
  }
  
  down_write(&current->mm->mmap_sem);
  spin_lock(&current->mm->page_table_lock);


  // first, get the current vma list and remove it
  newPrevVmaList = removeVmaRange(current->mm,GUEST_APP_LIMIT,&count);

  // get the data from the current page table for the previous address space
  for(i=0;i<NUM_PGD_GUEST;i++) {
    pgdout[i] = ((unsigned long *) current->mm->pgd)[i];
    ((unsigned long *) current->mm->pgd)[i] = pgdin[i];
  }
  
  // set up the new vma structs
  if(nextVma != NULL) {
    insertVmaList(current->mm, nextVma);
  }

  spin_unlock(&current->mm->page_table_lock);
  up_write(&current->mm->mmap_sem);

  // copy the data to the structure so it can be used later

  if(prevPgd != NULL) {
    // if a task is passed to us with data still on it, remove that
    // data before it is overwritten
    if(prevPgd[VMA_LIST_IDX] != 0) {
      //freeVmaList((struct vm_area_struct *) prevPgd[VMA_LIST_IDX]);
      printk("there should not be any existing data on the prevPgd\n");
      BUG();
    }

    // copy the new page table and vma information
    memcpy(&(prevPgd[VMA_LIST_IDX]),&newPrevVmaList,sizeof(unsigned long));
    memcpy(prevPgd,pgdout,NUM_PGD_GUEST*sizeof(unsigned long));
  } else {
    // there is no where to copy the data to, free the memory
    if(newPrevVmaList) {
      printk("sys_munmap did not free all guest app memory\n");
      BUG();
    }
    //freeVmaList(newPrevVmaList);
    //freePgdPages(pgdout);
  }

  __flush_tlb();
}

/**
 * this stores all of the os specific data associated with the pgd, but
 * does not free the pgd itself
 */
static void freePgd(unsigned long *pgd) {
  unsigned long *tmpPgd = (unsigned long *) get_free_page(GFP_KERNEL);
  do_switchguest(tmpPgd,pgd);
  sys_munmap(0,GUEST_APP_LIMIT);
  do_switchguest(pgd,tmpPgd);
  free_page((unsigned long) tmpPgd);

  // make sure everything got removed
  if(pgd != NULL) {
    if(pgd[VMA_LIST_IDX] != 0) {
      BUG();
    }
  }
}

asmlinkage int sys_switchguest(unsigned long prevIdx,unsigned long nextIdx) {
  unsigned long *prevPgd, *nextPgd;

  if(current->umlinux.guestPgdList == NULL) { BUG(); }
  if((prevIdx >= PTRS_PER_PAGE) || (nextIdx>=PTRS_PER_PAGE)) {
    BUG();
  }

  // get the storage
  prevPgd = (unsigned long *) current->umlinux.guestPgdList[prevIdx];
  nextPgd = (unsigned long *) current->umlinux.guestPgdList[nextIdx];

  //extract data from the storage structures
  if(prevPgd == nextPgd) {
    if(nextPgd != NULL) {
      return 0;
    }
  }

  // for the mysterious general protection fault
  // this should be fixed in the guest, and maybe someday I will.  How this
  // works is when the nextPgd is == NULL, then call munmap on the current
  // address space.  That way, there will be no data in the prevPgd data
  // structure and the guest kernel will dmenad mmap in the correct data.
  // I think this has something to do with vfork and/or clone, but am not
  // sure.  At this point I am just glad it works :) --STK 9/24/02
  if((nextPgd == NULL) || (prevPgd==NULL)) {
    sys_munmap(0,GUEST_APP_LIMIT);
  }

  // check to see if we are coming in with some data.  if so, free it
  if(prevPgd != NULL) {
    if(prevPgd[VMA_LIST_IDX] != 0) {
      freePgd(prevPgd);
    }
  }

  do_switchguest(prevPgd, nextPgd);

  return 0;
}



static int countVma(struct vm_area_struct *vma) {
  int count=0;
  while(vma != NULL) {
    vma = vma->vm_next;
    count++;
    if(count > 100000) {
      BUG();
    }
  }

  return count;
}

asmlinkage unsigned long sys_getguestpgd(int cmd, unsigned long pgdIdx) {
  unsigned long *newPgd;
  unsigned long retIdx;
  unsigned long *pgd=NULL;

  switch (cmd) {
  case 0:
    // search our current list and see if there is a space open
    if(current->umlinux.guestPgdList == NULL) { BUG(); }
    // start at 1 because we will always leave 0 as null
    for(retIdx=1;retIdx<PTRS_PER_PAGE;retIdx++) {
      if(current->umlinux.guestPgdList[retIdx] == 0) {
	break;
      }
    }
    if(retIdx>=PTRS_PER_PAGE) {BUG();}


    //new guest pgd, we want it zero'd out
    newPgd = (unsigned long *) get_free_page(GFP_KERNEL); 
    if(newPgd == NULL) {
      BUG();
    }
    current->umlinux.guestPgdList[retIdx] = (unsigned long) newPgd;

    return retIdx;
  case 1:
    // search our current list and see if there is a space open
    if((current->umlinux.guestPgdList == NULL) || (pgdIdx >= PTRS_PER_PAGE)) { 
      BUG(); 
    }
    
    pgd = (unsigned long *) current->umlinux.guestPgdList[pgdIdx];

    if(pgd != NULL) {      
      //count = freeVmaList((struct vm_area_struct *) pgd[VMA_LIST_IDX]);
      //freePgdPages(pgd);
      freePgd(pgd);

      free_page((unsigned long) pgd);  
      current->umlinux.guestPgdList[pgdIdx] = 0;
    }

    return 0;
  case 2:
    // this is for my debugging
    if(pgd == NULL) {
      current->thread.debugreg[7] = 0;
      asm("movl %0,%%db7": : "r" (0));
    } else {
      char printBuf[100];
      sprintf(printBuf,"debug addr 0x%08lx\n",(unsigned long) pgd);
      outputStr(printBuf);
      current->thread.debugreg[7] = 0x000d0303;
      current->thread.debugreg[0] = (unsigned long) pgd;
      asm("movl %0,%%db0": : "r" (current->thread.debugreg[0]));
      asm("movl %0,%%db7": : "r" (current->thread.debugreg[7]));
    }

    return 0;
  default:
    return 0;
  }

  //should not get here
  BUG();
  return 0;
	  
}


void freeGuestPgdList(unsigned long *list) {
  int i;
  unsigned long *pgd;

  if(list == NULL) {
    return;
  }

  
  for(i=1;i<PTRS_PER_PAGE;i++) {
    if(list[i] != 0) {
      pgd = (unsigned long *) list[i];
      //freePgdPages(pgd);
      //freeVmaList((struct vm_area_struct *) pgd[VMA_LIST_IDX]);
      freePgd(pgd);
      free_page((unsigned long) pgd);
    }
  }
  free_page((unsigned long) list);
}

EXPORT_SYMBOL_NOVERS(freeGuestPgdList);

asmlinkage int sys_syncprint(char *ubuf) {
  char kbuf[100];
  int ret = strncpy_from_user(kbuf,ubuf,99);
  if((ret>0) && (ret<100)) {
    kbuf[ret] = '\0';
    outputStr(kbuf);
  }

  return 0;
}
