/*
 * rlog.c - Hold log data in output buffer for reading from /proc/rlogd
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/smp_lock.h>
#include <linux/console.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/interrupt.h>			/* For in_interrupt() */

#include <asm/uaccess.h>

#include <revirt/rlog.h>

// Might be nice to make this a tunable kernel parameter
#define REPLAY_LOG_BUF_LEN	(1 << 22) /* This must be a power of two */
#define REPLAY_LOG_BUF_MASK	(REPLAY_LOG_BUF_LEN-1)

DECLARE_WAIT_QUEUE_HEAD(revirt_log_read_wait);
DECLARE_WAIT_QUEUE_HEAD(revirt_log_write_wait);

/*
 * Write protection for these should be provided by replay_logbuf_write_lock
 */
int replay_logwait=0;
struct task_struct * replay_waiting_tsk = NULL;

/*
 * replay_logbuf_lock protects replay_log_buf, replay_log_start, 
 * replay_log_end, replay_read_start
 * 
 * replay_logbuf_write_lock makes sure that writes are atomic; reads
 * may be interleaved with writes if necessary.
 */
static spinlock_t replay_logbuf_write_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t replay_logbuf_lock = SPIN_LOCK_UNLOCKED;

static char replay_log_buf[REPLAY_LOG_BUF_LEN];
#define REPLAY_LOG_BUF(idx) (replay_log_buf[(idx) & REPLAY_LOG_BUF_MASK])

/*
 * The indices into replay_log_buf are not constrained to REPLAY_LOG_BUF_LEN 
 * -- they must be masked before subscripting
 */
static unsigned long replay_log_start;	/* Index into replay_log_buf: next char to be read by replaylog() */
static unsigned long replay_lastread;	/* Index into replay_log_buf: beginning of last data read.  (Not safe to throw away)*/
static unsigned long replay_log_end;	/* Index into replay_log_buf: most-recently-written-char + 1 */

/*
 * Commands to do_replaylog:
 *
 * 	0 -- Close the log.  Currently a NOP.
 * 	1 -- Open the log. Currently a NOP.
 * 	2 -- Read from the log.
 *	9 -- Return number of unread characters in the log buffer
 * CLEANUP -- is this really necessary?  we don't have a replaylog syscall...
 */
int do_replaylog(int type, char * buf, int len)
{
	unsigned long i;
	char c;
	int error = 0;
	unsigned long read_end, real_start, real_end;

	switch (type) {
	case 0:		/* Close log */
		break;
	case 1:		/* Open log */
		break;
	case 2:		/* Read from log */
		error = -EINVAL;
		if (!buf || len < 0)
			goto out;
		error = 0;
		if (!len)
			goto out;
		error = verify_area(VERIFY_WRITE,buf,len);
		if (error)
			goto out;

		/*
		 * Note to self: waits until 2nd cond != 0 or signal pending
		 *
		 * Isn't this a race condition?
		 */
		wait_event_interruptible(revirt_log_read_wait, 
				    (replay_log_end - replay_log_start));
		if (error)
		      goto out;

		if(signal_pending(current)) {
			error=-ERESTARTSYS;
			goto out;
		}

		i = 0;

		replay_lastread = replay_log_start;
#if 0
		/*
		 * This works, but it's REEELY SLOW... try copy_user instead
		 * CLEANUP
		 */
		//printk("demonwrite: read about to acquire replay_logbuf_lock\n");
		spin_lock_irq(&replay_logbuf_lock);
		while ((replay_log_start != replay_log_end) && i < len) {
			c = REPLAY_LOG_BUF(replay_log_start);
			replay_log_start++;
			//printk("demonwrite: releasing replay_logbuf_lock\n");
			spin_unlock_irq(&replay_logbuf_lock);
			__put_user(c,buf);
			buf++;
			i++;
			//printk("demonwrite: read about to acquire replay_logbuf_lock 2\n");
			spin_lock_irq(&replay_logbuf_lock);
		}
#endif
		spin_lock_irq(&replay_logbuf_lock);

		if(replay_log_start > replay_log_end) {
			printk("replaylog: end before start: [%x,%x)\n",
				replay_log_start, replay_log_end);
			BUG();
		}
		/*
		 * Find the last "virtual index": either fill
		 * request, or whatever we have
		 */
		read_end = replay_log_start + len;
		if(read_end > replay_log_end)
			read_end = replay_log_end;

		/*
		 * Now find the actual index from the virtual indices...
		 */

		real_start = replay_log_start & REPLAY_LOG_BUF_MASK;
		real_end = read_end & REPLAY_LOG_BUF_MASK;

	//	printk("replaylog: start:0x%x end:0x%x readend:0x%x real_s:%x real_e:%x\n", 
	//		replay_log_start, replay_log_end, read_end,
	//		real_start, real_end);
		/*
		 * If it's contiguious, do one copy; else, do
		 * two copies.
		 */
		if(real_end > real_start) {
			copy_to_user(buf, 
				 replay_log_buf + real_start, 
				 real_end-real_start);
		} else {
			unsigned long d1, s1, l1, d2, s2, l2;
			s1 = (unsigned long)replay_log_buf + real_start;
			l1 = REPLAY_LOG_BUF_LEN - real_start;
			s2 = (unsigned long)replay_log_buf;
			l2 = real_end;
			d1 = (unsigned long)buf;
			d2 = (unsigned long)buf + l1;
		//	printk("replaylog: (%x,%x,%x) (%x,%x,%x)\n", 
		//		d1, s1, l1, d2, s2, l2);
			copy_to_user((void *)d1, (void *)s1, l1);
			copy_to_user((void *)d2, (void *)s2, l2);
		}
		i=read_end - replay_log_start;
		replay_log_start = read_end;
		spin_unlock_irq(&replay_logbuf_lock);
		
		//printk("demonwrite: wake_up(revirt_log_write_wait)\n");
		wake_up(&revirt_log_write_wait);
		//printk("demonwrite: releasing replay_logbuf_lock 2\n");
//		spin_unlock_irq(&replay_logbuf_lock);
		error = i;
		break;
	case 9:		/* Number of chars in the log buffer */
		//printk("demonwrite: poll about to acquire replay_logbuf_lock\n");
		spin_lock_irq(&replay_logbuf_lock);
		error = replay_log_end - replay_log_start;
#if 0
//		printk("replaylog(POLL):log_end %x log_start %x\n",
				replay_log_end, replay_log_start);
#endif
		//printk("demonwrite: releasing replay_logbuf_lock 2\n");
		spin_unlock_irq(&replay_logbuf_lock);
		break;
	default:
		error = -EINVAL;
		break;
	}
out:
	return error;
}

#if 0
asmlinkage long sys_replaylog(int type, char * buf, int len)
{
	if ((type != 3) && !capable(CAP_SYS_ADMIN))
		return -EPERM;
	return do_replaylog(type, buf, len);
}
#endif

static void emit_rlog_char(char c)
{
	int flag;
	//printk("demonwrite: emit_rlog_char about to acquire replay_logbuf_lock\n");
	spin_lock_irqsave(&replay_logbuf_lock, flag);
	REPLAY_LOG_BUF(replay_log_end) = c;
	replay_log_end++;
	/*
	 * Now that we're dealing with gigabytes, the log indices
	 * may overflow.  Rather than trying to deal with 64-bit indices,
	 * we detect overflow and compensate.  This means we need to be
	 * very careful about renewing our information before unlocking...
	 */
	if(replay_log_end & 0x80000000) {
		replay_log_end = (replay_log_end & REPLAY_LOG_BUF_MASK) +
					REPLAY_LOG_BUF_LEN;
		replay_log_start = replay_log_start & REPLAY_LOG_BUF_MASK;
		replay_lastread = replay_lastread & REPLAY_LOG_BUF_MASK;
		printk("rlog: index overflows, reset to [%x,%x)\n",
				replay_log_start, replay_log_end);
	}


	//printk("demonwrite: releasing replay_logbuf_lock 3\n");
	spin_unlock_irqrestore(&replay_logbuf_lock, flag);
	/*
	 * This condition should never happen; the write lock combined with 
	 * the condition wait should guarantee that.
	 */
	if (replay_log_end - replay_log_start > REPLAY_LOG_BUF_LEN) {
	   printk(KERN_EMERG "Replay Log overflow!!! Attempting to call schedule...\n");
	   schedule();
	}
	/*
	 * Just because we've sent the data to the rlogd process doesn't mean
	 * it's safely on disk.  We assume, however, that when rlogd calls
	 * read(/proc/rlog) again, all previously read data is on the disk.
	 * This warning detects the condition of overwriting data that has
	 * been sent to rlogd but we're not sure had made it to the disk yet.
	 */
	if (replay_log_end - replay_lastread > REPLAY_LOG_BUF_LEN) {
	   printk("WARNING: Overwriting unsync'd log data\n");
	   replay_lastread=replay_log_start;
	}
}

/*
 * This is demon_log_write. It can be called from any context.  
 * We want it to work.
 * We want to guarantee atomic writes of records, so this should be
 * with an entire record to write.
 */
asmlinkage int demon_write_log(char * buf, int type, int len) {
	struct iovec v = { buf, len };
	return demon_writev_log(type, &v, 1);
}

/*
 * If a logged process is blocked waiting for rlogd to empty the buffer,
 * we must not schedule another logged process.
 */
asmlinkage int replay_log_schedule(void) {
	return !replay_logwait;
}

int demon_flush_if_full(int type) { 
	if((replay_log_end - replay_log_start > REPLAY_LOG_BUF_LEN / 4)
		 || (type == ENTRY_EXIT)) {
		//printk("demonwrite: about to wake up revirt_log_read_wait\n");
		wake_up_interruptible(&revirt_log_read_wait);
	}
}

asmlinkage int demon_writev_log(int type, const struct iovec * v, int n)
{
	static int ccount=0;
	int inst=ccount++;
	unsigned long flags;
	int printed_len=0, i,j,  len, error;
	char *p;
	struct rlog_header h;

	for(len=j=0; j<n; j++)
		len+=v[j].iov_len;

	h.type=type;
	h.len=len;

	/*
	 * This is from the printk code. Is this necessary?  
	 */
	if (oops_in_progress) {
		/* If a crash is occurring, make sure we can't deadlock */
		spin_lock_init(&replay_logbuf_write_lock);
	}

	
	len+=sizeof(h);

	/*
	 * In fact, if we did the space checking around emit_char() instead of
	 * here, we could handle lengths larger than the buffer size;
	 * per'aps we can add that later, but as we plan on the log being
	 * in the order of megabytes, it't not important right now.
	 */
	if(len > REPLAY_LOG_BUF_LEN) {
		printk("demonwrite: tried to write log of size %d, limit %d\n",
			len, REPLAY_LOG_BUF_LEN);
		while(1);
	}

	/* FIXME -- hey, these aren't the right locks... we need semaphores */
	spin_lock_irqsave(&replay_logbuf_write_lock, flags);

	while(replay_log_end - replay_log_start + len > REPLAY_LOG_BUF_LEN) {
	    int error;
	    replay_logwait=1;
	    replay_waiting_tsk = current;
	    printk("demonwrite: waiting for enough room to write data: s:%x e:%x l:%x LEN:%x\n", replay_log_start, replay_log_end, len, REPLAY_LOG_BUF_LEN);
	    wait_event(revirt_log_write_wait, 
	        (!(replay_log_end - replay_log_start + len > REPLAY_LOG_BUF_LEN)));
	    replay_logwait=0;
	    replay_waiting_tsk = NULL;
	}

	for(p= (char *)&h, i=0; i<sizeof(h); i++, p++)  {
		emit_rlog_char(*p);
		printed_len++;
	}		

	for(j=0; j<n; j++) {
	    p=v[j].iov_base;
	    for (i=0; i<v[j].iov_len; i++) {
		emit_rlog_char(*p);
		p++;
		printed_len++;
	    }
	}

	spin_unlock_irqrestore(&replay_logbuf_write_lock, flags);

	/*
	 * For optimization, only wake up rlogd when log has reached some
	 * watermark, and when a logged process exits.  We don't want to
	 * schedule after a signal, however, so don't wake up then.
	 *
	 * Note, this means that REPLAY_SCHED should never be set when
	 * writing type ENTRY_EXIT.
	 */
	if((printed_len && !(current->replay_flags & REPLAY_SCHED))) 
		demon_flush_if_full(type);
	return len;
}
EXPORT_SYMBOL(demon_write_log);
