/*
 * umlinux.c - umlinux kernel mode VMM
 *
 * Copyright (C) 2002 Sam King
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * this is the kernel module used to emulate the UMLinux ptrace
 * tracer process.  It registers callback functions with the
 * kernel so that it is notified every time one of the following
 * events occurs:
 *   1) UMLinux made a system call and it is about to be passed
 *      to the kernel system call handler
 *   2) a system call made by UMLinux just returned
 *   3) a signal is about to be sent to the UMlinux process
 *
 * For specific on how the tracer process is supposed to work
 * read through the source code.  This is a work in progress
 * and will change.
 *
 * @author Sam King
 * @date 3/15/02
 * @version 1.0
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <linux/ptrace.h>
#include <asm/ptrace.h>
#include <asm/signal.h>
#include <asm/io.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/desc.h>

//#define SNIFF_INTERFACE
//#define COUNT_SYS
//#define EMULATE_RDTSC

#ifdef SNIFF_INTERFACE
extern int stkprint;
static int guestRet=0;
static int hostRet=0;
static char printBuf[100];
#endif

#define USE_MPROTECT 0
#define USE_MMAP 0
#define PROTECT_KERNEL 1

//#undef outputStr
//#define outputStr(s) printk(s)

/**
 * this was copied from the lilo.h UMLinux header file.  It is the address
 * where the guest kernel begins.
 */
#define MEM_BASE (0x70000000)
#define MY_PROT  (VM_READ | VM_WRITE | VM_EXEC)
/**
 * these should be in a header file, but were cut and pasted from
 * signal.c and ptrace.c (both in arch/i386/kernel).  These define
 * the callback function signature
 */
typedef int (*syscall_callback_sig)(struct pt_regs *regs);
typedef int (*signal_callback_sig)(struct pt_regs *regs, int signr);

/**
 * this is a small helper function used to print the registers
 */
static void printkRegisters(struct pt_regs *regs) 
{
  printk("ebx 0x%lx\n",regs->ebx);
  printk("ecx 0x%lx\n",regs->ecx);
  printk("edx 0x%lx\n",regs->edx);
  printk("esi 0x%lx\n",regs->esi);
  printk("edi 0x%lx\n",regs->edi);
  printk("ebp 0x%lx\n",regs->ebp);
  printk("eax %ld\n",regs->eax);
  printk("xds %d\n",regs->xds);
  printk("xes %d\n",regs->xes);
  printk("orig_eax %ld\n",regs->orig_eax);
  printk("eip 0x%lx\n",regs->eip);
  printk("xcs %d\n",regs->xcs);
  printk("eflags 0x%lx\n",regs->eflags);
  printk("esp 0x%lx\n",regs->esp);
  printk("xss %d\n",regs->xss);
}

/**
 * the functions exported by signal.c and ptrace.c, repsectivly
 */
extern void umlinux_reg_signal_callback(signal_callback_sig callback);
extern int umlinux_reg_syscall_callback(syscall_callback_sig callback);
extern asmlinkage long sys_mprotect(unsigned long start, size_t len, unsigned long prot);
//unsigned long do_mmap_pgoff(struct file * file, unsigned long addr, unsigned long len,
//unsigned long prot, unsigned long flags, unsigned long pgoff)
extern asmlinkage long sys_mmap2(unsigned long addr, unsigned long len,
		 unsigned long prot, unsigned long flags,
		 unsigned long fd, unsigned long pgoff);
extern asmlinkage long sys_munmap(unsigned long addr, size_t len);
extern void freeGuestPgdList(unsigned long *list);

/**
 * here are some local functions that may get used
 */
static int getUmlinuxKernelSize(void);
static void umlkernel_mmap();
static void umlkernel_munmap();

#define PAUSEMASK (0xff)
static int mmap_count=0;

extern void ztrace_debug_pause(void);
extern void ztrace_yield_to(struct task_struct * );

#ifdef COUNT_SYS
static int countSys=0;
static unsigned long numHostSys=0;
static unsigned long numGuestSys=0;
static unsigned long numSigs[100];
static unsigned long numSwitch=0;
static unsigned long numCtx=0;
#endif

/**
 * this is the function that will get called every time a system
 * call either a) is about to be handled, or b) just returned and
 * control is about to be returned back to the guest.
 *
 * @param regs this is a pointer to the guest process registers
 *        WARNING: this is a pointer to the actual stack being used
 *        so know if you change something here, it will really 
 *        be changed (it is not a copy).
 *
 * @return 0, process the call as normal or SIG something for
 *         that signal to be sent to the guest process before
 *         control is returned
 */
static int syscall_callback(struct pt_regs *regs)
{
  int cont=0;

  switch (current->umlinux.syscallState) {
  case UML_SYSCALL_ENTER: // the system call has not been called yet

    current->umlinux.syscallState = UML_SYSCALL_RETURN;
    // check if it should be sent to the kernel, or redirected to the guest
    // through a signal

    if( (current->umlinux.umlinuxMode == UML_KERNEL_MODE) &&
        ( (regs->orig_eax ==0) || 
	  (regs->orig_eax == __NR_rt_sigreturn && MEM_BASE <= regs->eip) ) ) {

#ifdef SNIFF_INTERFACE
      if(stkprint) {
	if(regs->orig_eax == 0) {
	  hostRet=1;
	  if((regs->edi & 0xffff0000) == 0xcafe0000) {
	    sprintf(printBuf,"host 0x%08lx %lu\n",
		    regs->edi,regs->edi & 0x0000ffff);
	  } else {	    
	    sprintf(printBuf,"host 0x%08lx %lu\n",
		    regs->edi,regs->edi >> 16);
	  }
	  outputStr(printBuf);  
	}
      }
#endif
      

      // send it to the host kernel
      current->umlinux.syscallType = UML_SYSTEM_SYSCALL;

      // reset the registers for the host system call
      if (regs->orig_eax == 0 && (regs->edi & 0xffff0000) == 0xcafe0000) {
	// it is the old cafe style of lotol system call
	regs->orig_eax = regs->edi & 0x0000ffff;
      } else if (regs->orig_eax == 0) {
	// this is for the new style with lotol__llseek support
	regs->orig_eax = regs->edi >> 16;
	regs->edi &= 0x0000ffff;
      }

#ifdef COUNT_SYS
      if(countSys) {
	numHostSys++;
	if(regs->orig_eax == __NR_switchguest) {
	  numCtx++;
	}
      }
#endif

    
      // check if it is an exit system call
      if (regs->orig_eax == __NR_exit) {
	// cleanup
	//freeGuestPgdList(current->umlinux.guestPgdList);
	//current->umlinux.guestPgdList = NULL;
	cont = 0;
      } else if (regs->orig_eax == __NR_rt_sigreturn) {
	// it is a regreturn from the guest kernel (we hope)
	current->umlinux.syscallReturnType = UML_RT_SIGRETURN;

	cont = 0;
      } else  {
	// it is a system call that should go through
	cont = 0;
      }
      
    } else {
      // the system call should be sent back to the guest kernel
      current->umlinux.syscallType = UML_UMLINUX_SYSCALL;

#ifdef COUNT_SYS
      if(countSys) {numGuestSys++;}
#endif

#ifdef SNIFF_INTERFACE
      if(stkprint) {
	guestRet=1;
	sprintf(printBuf,"guest %lu\n",regs->orig_eax);
	outputStr(printBuf);
      }
      
      if(regs->orig_eax == __NR_modify_ldt) {
	stkprint = (stkprint) ? 0 : 1;
      }
#endif

      
      /*
      {
	struct timeval tv;
	do_gettimeofday(&tv);
	if(regs->orig_eax == __NR_modify_ldt) {
	  printk("time: %ld.%06ld\n",tv.tv_sec,tv.tv_usec);
	}
      }
      */

#ifdef COUNT_SYS
      if(regs->orig_eax == __NR_modify_ldt) {
	if(countSys) {
	  countSys=0;
	  printk("num host syscalls: %lu\n",numHostSys);
	  printk("num guest syscalls: %lu\n",numGuestSys);
	  printk("num switches: %lu\n",numSwitch);
	  printk("num ctx: %lu\n",numCtx);
	  {
	    int i;
	    for(i=0;i<100;i++) {
	      if(numSigs[i]>0) {
		printk("num sig %d: %lu\n",i,numSigs[i]);
	      }
	    }
	  }
	} else {
	  int i;
	  for(i=0;i<100;i++) {
	    numSigs[i] = 0;
	  }
	  numHostSys=0;
	  numGuestSys=0;
	  numSwitch=0;
	  numCtx=0;
	  countSys=1;
	}
      }
#endif
      

      // fake out the kernel so that it gets a getpid.  This could
      // also be a -ENOSYS for a null system call, but it does not matter
      // since getpid only changes eax
      current->umlinux.saved_orig_eax = regs->orig_eax;
      regs->orig_eax = 20;   // for getpid
      cont = SIGUSR1;  // notify the guest kernel with a SIGUSR1
    }

    break;

  case UML_SYSCALL_RETURN: // the system call has been processed, 
                           // it is about to return to the guest
                       

    current->umlinux.syscallState = UML_SYSCALL_ENTER;
    if(current->umlinux.syscallType == UML_UMLINUX_SYSCALL) {
      // the system call was meant for the umlinux kernel, restore the
      // original registers
      regs->eax = current->umlinux.saved_orig_eax;
      regs->orig_eax = current->umlinux.saved_orig_eax;
      
    } else if(current->umlinux.syscallReturnType == UML_RT_SIGRETURN) {
      current->umlinux.syscallReturnType = UML_NOT_RT_SIGRETURN;
      if((regs->eip < MEM_BASE) && (current->umlinux.umlinuxMode == UML_KERNEL_MODE)) {

#ifdef SNIFF_INTERFACE
	if(guestRet) {
	  guestRet=0;
	  sprintf(printBuf,"gr 0x%08lx\n",regs->eax);
	  outputStr(printBuf);
	}
#endif

	//making the switch
	umlkernel_munmap();
	current->umlinux.umlinuxMode = UML_USER_MODE;
      }
    } else {
      
#ifdef SNIFF_INTERFACE
      if(hostRet) {
	hostRet=0;
	sprintf(printBuf,"hr 0x%08lx %ld\n",regs->eax,(long) regs->eax);
	outputStr(printBuf);
      }
#endif

    }

    break;
  }

  return cont;
}

/**
 * this is the callback function that is called right before
 * a signal is passed to the guest process.  you can change
 * the signal by returning something other than signr
 *
 * @param regs this is a pointer to the guest process registers
 *        WARNING: this is a pointer to the actual stack being used
 *        so know if you change something here, it will really 
 *        be changed (it is not a copy).
 *
 * @param signr the signal number.
 * @return I am not sure what 0 will do, but if you return a valid signal
 *         number that is the signal that will be sent.  
 */
static int signal_callback(struct pt_regs *regs, int signr)
{
  int ret = signr;

#ifdef EMULATE_RDTSC
  unsigned long opcode;
  unsigned long low=0, high=0;

  copy_from_user(&opcode,(unsigned long *) regs->eip,sizeof(unsigned long));
  if((signr == SIGSEGV) && ((opcode & 0x0000ffff) == 0x0000310f)) {
    //we have an rdtsc instruction, emulate it
    __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high));
    regs->eax = low;
    regs->edx = high;
    regs->eip += 2;
    
    ret = 0;
  }
#endif

#ifdef SNIFF_INTERFACE
      if(stkprint) {
	char mybuf[256];
#ifdef EMULATED_RDTSC
	sprintf(mybuf,"signal %d opcode 0x%08lx\n",signr,opcode);
#else
	sprintf(mybuf,"signal %d\n",signr);
#endif
	outputStr(mybuf);

      }
#endif

  if(signr == SIGTRAP) {
    // this is from an int3
    ret = SIGUSR2;
  }

  // if we are in user mode, we are about to switch back to 
  // kernel mode
  if(current->umlinux.umlinuxMode == UML_USER_MODE) {
    current->umlinux.umlinuxMode = UML_KERNEL_MODE;
    umlkernel_mmap();
  }

#ifdef COUNT_SYS
  if(countSys) { if(signr>0){numSigs[signr]++;} }
#endif

  return ret;
}

/**
 * this function is used to calculate how much memory was mmap'd for
 * the umlinux kernel
 */
static int getUmlinuxKernelSize(void) {
  struct vm_area_struct *vma;
  int found = 0,i;

  vma = current->mm->mmap;
  for(i=0;i<current->mm->map_count;i++) {
    if(vma->vm_start == MEM_BASE) {
      found=1;
      break;
    }

    vma = vma->vm_next;
  }

  if(!found) {return -1;}

  //check if the next one is mem-base + pagesize
  if((vma->vm_next != NULL) && 
     (vma->vm_next->vm_start == (MEM_BASE+PAGE_SIZE))) {

    return vma->vm_next->vm_end - vma->vm_next->vm_start +
      PAGE_SIZE;
  }
  
  return vma->vm_end - vma->vm_start;
}

static void umlkernel_mmap(struct pt_regs *regs) {
#if PROTECT_KERNEL
# if USE_MPROTECT
  int len;
  long int ret;

  // this is the mprotect version
  len = getUmlinuxKernelSize() - PAGE_SIZE;

  if(len<=0) {
    printk("get umlinux kerne size returned %d\n",len);
  } else {
    if(sys_mprotect(MEM_BASE+PAGE_SIZE,len,PROT_READ | PROT_WRITE | PROT_EXEC)) {
      printk("mprotect failed\n");
    }
  }
  
# elif USE_MMAP
  int len;
  long int ret;

  // this is the mmap version
  //len = getUmlinuxKernelSize();
  len = current->umlinux.guestKernelSize;
  if(len<=0) {
    printk("get umlinux kerne size returned %d\n",len);
  } else {
    ret = sys_mmap2(MEM_BASE+PAGE_SIZE, len,PROT_READ | PROT_WRITE | PROT_EXEC,                   
    			MAP_SHARED | MAP_FIXED, 0x10, PAGE_SIZE >> PAGE_SHIFT);
    if(ret != (MEM_BASE+PAGE_SIZE)) {
      printk("mmap failed %ld\n",ret);
    }
  }
# else  
//  printk("umlkernel_mmap: setting kernel prot for pid %d [pause]\n", current->pid);
  if(!(mmap_count & PAUSEMASK)) {
//	ztrace_yield_to(NULL);
  }
  current->umlinux.limit = 0xffffffff;
  um_set_limit(__USER_CS >> 3, current->umlinux.limit);
  um_set_limit(__USER_DS >> 3, current->umlinux.limit);
//  printk("umlkernel_mmap: done. [pause]\n");
  if(!(mmap_count & PAUSEMASK)) {
//	ztrace_yield_to(NULL);
  }
  mmap_count++;

# endif
#endif 
}       

static void umlkernel_munmap(struct pt_regs *regs) {
#if PROTECT_KERNEL
# if USE_MPROTECT
  int len;
  int ret;

  len = getUmlinuxKernelSize() - PAGE_SIZE;

  if(len<=0) {
    printk("get umlinux kernel size returned %d\n",len);
  } else {
    if(sys_mprotect(MEM_BASE+PAGE_SIZE,len,PROT_NONE)) {
      printk("mprotect2 failed\n");
    }
  }
# elif USE_MMAP
  int len;
  int ret;

  // this is the mmap version
  len = getUmlinuxKernelSize() - PAGE_SIZE;

  current->umlinux.guestKernelSize = len;
  if(len<=0) {
    printk("get umlinux kernel size returned %d\n",len);
  } else {
    ret = sys_munmap(MEM_BASE+PAGE_SIZE,len);
    if(ret != 0) {
      printk("munmap failed\n");
    }
  }
# else

#ifdef COUNT_SYS
  if(countSys) {numSwitch++;}
#endif

//  printk("umlkernel_munmap: clearing kernel prot for pid %d [pause]\n", current->pid);
  if(!(mmap_count & PAUSEMASK)) {
//	ztrace_yield_to(NULL);
  }
//  current->umlinux.limit = 0x6fffffff;
  current->umlinux.limit = 0x6fffffff;
  um_set_limit(__USER_CS >> 3, current->umlinux.limit);
  um_set_limit(__USER_DS >> 3, current->umlinux.limit);
//  printk("umlkernel_munmap: done. [pause]\n");
  if(!(mmap_count & PAUSEMASK)) {
//	ztrace_yield_to(NULL);
  }
    
# endif
#endif

}

/**
 * linux module stuff
 */
int init_module(void) 
{
  /**
   * regiter for the callbacks
   */
  umlinux_reg_syscall_callback(syscall_callback);
  umlinux_reg_signal_callback(signal_callback);


  return 0;
}

void cleanup_module(void)
{
  /**
   * make sure you nullify the callbacks, the kernel will (probably) crash
   * if you do not.
   */
  umlinux_reg_syscall_callback(NULL);
  umlinux_reg_signal_callback(NULL);
}

